from __future__ import print_function

import epics
import time
import pvnames
pvname1 = pvnames.updating_pvlist[0]
pvname2 = pvnames.updating_pvlist[1]

def wait(t=30):
    t0 = time.time()
    while time.time()-t0 < t:
        time.sleep(0.01)

def onChange(pvname,  value=None, char_value=None, timestamp=None, **kw):
    print('  new value: %s = %s (%s) ' % ( pvname, char_value, time.ctime(timestamp)))

epics.camonitor(pvname1, callback=onChange)
epics.camonitor(pvname2, callback=onChange)


print('## Monitor 2 PVs with epics.camonitor for 10sec')

wait(10)

print('## clear monitor for ', pvname2)

epics.camonitor_clear(pvname2)

print('## Monitor remaining PV for 10sec')
wait(10)

print('done!')



#!/usr/bin/env python
"""
String Utils for Python 2
"""
import sys
if sys.version_info[0] != 2:
    raise ImportError(" Python version 2 required")

NULLCHAR_2 = '\x00'
NULLCHAR =  '\x00'
STR2BYTES =  str
BYTES2STR = str

def strjoin(sep, seq):
    "join string sequence with a separator"
    return sep.join(seq)

def is_string(s):
    return isinstance(s, basestring)

is_string_or_bytes = is_string

ascii_string = str
# def ascii_string(s):
#     if isinstance(s, unicode):
#         return str(s)
#     else:
#         return s



#!/usr/bin/env python
"""
String Utils for Python 3
"""
import os
import sys
if sys.version_info[0] != 3:
    raise ImportError(" Python version 3 required")

EPICS_STR_ENCODING = os.environ.get('PYTHONIOENCODING', 'utf-8')
NULLCHAR_2 = '\x00'
NULLCHAR   = b'\x00'

def s2b(st1):
    'string to byte conversion'
    if isinstance(st1, bytes):
        return st1
    return bytes(st1, EPICS_STR_ENCODING)

def b2s(st1):
    'byte to string conversion'
    if isinstance(st1, str):
        return st1
    elif isinstance(st1, bytes):
        return str(st1, EPICS_STR_ENCODING)
    else:
        return str(st1)

STR2BYTES, BYTES2STR = s2b, b2s

def strjoin(sep, seq):
    "join string sequence with a separator"
    if isinstance(sep, bytes):
        sep = BYTES2STR(sep)
    if len(seq) == 0:
        seq = ''
    elif isinstance(seq[0], bytes):
        tmp =[]
        for i in seq:
            if i == NULLCHAR:
                break
            tmp.append(BYTES2STR(i))
        seq = tmp
    return sep.join(seq)

def is_string(s):
    return isinstance(s, str)

def is_string_or_bytes(s):
    return isinstance(s, str) or isinstance(s, bytes)

def ascii_string(s):
    return bytes(str(s), EPICS_STR_ENCODING)



from __future__ import print_function
import time
import epics 
import pvnames

p = epics.PV(pvnames.updating_pv1, auto_monitor= False)

def onChange(pvname=None,char_value=None,value=None,**kw):
    print(pvname, value, time.ctime())
p.add_callback(onChange)

t0 = time.time()
p.get()
while time.time()-t0 < 20:
    print(p.get())
    time.sleep(0.1)
    



import time
import epics
import pvnames
import random
import numpy
import sys
driver = epics.PV(pvnames.subarr_driver)
sub1   = epics.PV(pvnames.subarr1)
sub2   = epics.PV(pvnames.subarr2)
sub3   = epics.PV(pvnames.subarr3)
sub4   = epics.PV(pvnames.subarr4)


s1_0 =  int(epics.caget("%s.INDX" % pvnames.subarr1))
s2_0 =  int(epics.caget("%s.INDX" % pvnames.subarr2))
s3_0 =  int(epics.caget("%s.INDX" % pvnames.subarr3))
s4_0 =  int(epics.caget("%s.INDX" % pvnames.subarr4))

s1_n =  int(epics.caget("%s.NELM" % pvnames.subarr1))
s2_n =  int(epics.caget("%s.NELM" % pvnames.subarr2))
s3_n =  int(epics.caget("%s.NELM" % pvnames.subarr3))
s4_n =  int(epics.caget("%s.NELM" % pvnames.subarr4))

npts = len(driver.get())

for i in range(10):
    driver.put([100*random.random() for x in range(npts)])
    time.sleep(0.1)

    full = driver.get()
    sys.stdout.write("%s\n" % full)
    all_ok = all( [all( sub1.get() == full[s1_0:s1_n+s1_0]), 
                  all( sub2.get() == full[s2_0:s2_n+s2_0]),
                  all( sub3.get() == full[s3_0:s3_n+s3_0]),
                  all( sub4.get() == full[s4_0:s4_n+s4_0])
                  ])

    if not all_ok:
        sys.stdout.write("FAIL: Subarrays don't match larger array!\n")
        sys.exit()
        
sys.stdout.write('PASSED.\n')
    



import sys
import time
import epics
import pvnames

HAS_NUMPY = False
try:
    import numpy
    HAS_NUMPY = True
except ImportError:
    pass
   
pvlist = (
    pvnames.str_pv,
    pvnames.int_pv,
    pvnames.float_pv,       
    pvnames.enum_pv,
    pvnames.char_arr_pv,
    pvnames.long_pv,
    pvnames.long_arr_pv,       
    pvnames.double_pv,
    pvnames.double_arr_pv,
    pvnames.string_arr_pv,
    )

def onConnect(pvname=None,  **kw):
    print(' on Connect %s --  %s' % (pvname, repr(kw)))
    
def onChanges(pvname=None, value=None, **kw):
    print(' on Change %s =  %s' % (pvname, repr(value)))
        

def RunTest(pvlist, use_preempt=True, maxlen=16384, 
            use_numpy=True, form='native'):
    msg= ">>>Run Test: %i pvs, numpy=%s, form=%s, preempt=%s"
    print( msg % (len(pvlist), use_numpy, form, use_preempt))

    epics.ca.HAS_NUMPY = use_numpy and HAS_NUMPY
    epics.ca.PREEMPTIVE_CALLBACK = use_preempt
    epics.ca.AUTOMONITOR_MAXLENGTH = maxlen
    mypvs= []
    for pvname in pvlist:
        pv = epics.PV(pvname, form=form,
                      # connection_callback=onConnect,
                      # callback=onChanges
                      )
        mypvs.append(pv)
    epics.poll(evt=0.10, iot=10.0)

    for pv in mypvs:
        # time.sleep(0.1)
        # epics.poll(evt=0.01, iot=1.0)
        val  = pv.get()
        cval = pv.get(as_string=True)    
        if pv.count > 1:
            val = val[:12]
        print( '-> ', pv, cval)
        print( '   ', type(val), val)
    for pv in mypvs:
        pv.disconnect()
    time.sleep(0.01)


for use_preempt in (True, False):
    for use_numpy in (False,):
        for form in ('native', 'time', 'ctrl'):
            time.sleep(0.001)
            RunTest(pvlist,
                    use_preempt=use_preempt,
                    use_numpy=use_numpy,
                    form=form)
        # sys.exit()




#!/usr/bin/python

# test of simplest device

from epics.devices import ao
import sys
import time

myao = ao('Py:ao1')


if len(myao._pvs) < 10:
    print(" Not enough PVS!!")
    sys.exit()

if myao.VAL is None:
    print(" Value is None!!")
    sys.exit()


myao.write_state('tmp_aostate.txt')

time.sleep(0.5)

flines = open('tmp_aostate.txt').readlines()
if len(flines) < 10:
    print(" write_state didn't work properly")
    sys.exit()

print(" all tests passed!")





from ctypes import cdll
from epics.ca import find_libca
dllname = find_libca()

print(dllname)

libca = cdll.LoadLibrary(dllname)

print(hasattr(libca, 'ca_context_create'),
      hasattr(cdll.LoadLibrary(None), 'ca_context_create'))



import time
import epics
import sys
import pvnames

# Test than when a PV connections all callbacks fire successfully
#
# does not require Setup/simulator.py to be running (PV is deliberately one which does not change)

write = sys.stdout.write

got_callback_a = False
got_callback_b = False

def callback_a(pvname=None, value=None, **kw):
    global got_callback_a
    write( "Got callback A (%s, %s)\n" % (pvname, repr(value)) )
    got_callback_a = True

def callback_b(pvname=None, value=None, **kw):
    global got_callback_b
    write( "Got callback B (%s, %s)\n" % (pvname, repr(value)) )
    got_callback_b = True


pvname = pvnames.non_updating_pv
mypv = epics.PV(pvname, callback=(callback_a, callback_b))

write('Created PV with two callbacks = %s\n' % mypv)

write('Now wait for changes...\n')

def wait(timeout=10):
    t0 = time.time()
    while time.time() - t0 < timeout and not (got_callback_a and got_callback_b):
        time.sleep(1.e-4)

wait(2)

if not mypv.connected:
    write('ERROR: PV never connected\n')
    sys.exit(1)

if not (got_callback_a and got_callback_b):
    write('ERROR: Inconsistent initial value callbacks - callback A = %s, callback B = %s\n'
          % (got_callback_a, got_callback_b) )
    sys.exit(1)

write('Got both callbacks OK!\n')



from epics import ca

import time
import sys
import pvnames
mypv = pvnames.updating_pv1

def onChanges(pvname=None, value=None, **kw):
    sys.stdout.write( 'New Value: %s  value=%s, kw=%s\n' %( pvname, str(value), repr(kw)))
    sys.stdout.flush()
    
chid = ca.create_channel(mypv)
eventID = ca.create_subscription(chid, callback=onChanges)

t0 = time.time()
while time.time()-t0 < 15.0:
    time.sleep(0.001)






import time
import sys

from epics import ca

import pvnames

pvname = pvnames.updating_pv1

def wait(step=0.1, maxtime=30):
    t0 = time.time()
    while time.time()-t0 < maxtime:
        time.sleep(step)

    
def setup_callback(pvname):
    def my_cb(pvname=None, value=None, **kw):
       sys.stdout.write( 'get: %s  value=%s, kw=%s\n' %( pvname, str(value), repr(kw)))
       sys.stdout.flush()

    chid = ca.create_channel(pvname)
    return ca.create_subscription(chid, callback=my_cb)

cb_ref = setup_callback(pvname)

wait()
    




from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

__doc__ = """
   epics channel access python module

   version: %s
   Principal Authors:
      Matthew Newville <newville@cars.uchicago.edu> CARS, University of Chicago
      Angus Gratton <angus.gratton@anu.edu.au>, Australian National University

== License:

   Except where explicitly noted, this file and all files in this
   distribution are licensed under the Epics Open License See license.txt in
   the top-level directory of this distribution.

== Overview:
   Python Interface to the Epics Channel Access
   protocol of the Epics control system.

""" % (__version__)


import time
import sys
import threading
from . import ca
from . import dbr
from . import pv
from . import alarm
from . import device
from . import motor
from . import multiproc

PV    = pv.PV
Alarm = alarm.Alarm
Motor = motor.Motor
Device = device.Device
poll  = ca.poll

get_pv = pv.get_pv

CAProcess = multiproc.CAProcess
CAPool = multiproc.CAPool

# some constants
NO_ALARM = 0
MINOR_ALARM = 1
MAJOR_ALARM = 2
INVALID_ALARM = 3

_PVmonitors_ = {}

def caput(pvname, value, wait=False, timeout=60):
    """caput(pvname, value, wait=False, timeout=60)
    simple put to a pv's value.
       >>> caput('xx.VAL',3.0)

    to wait for pv to complete processing, use 'wait=True':
       >>> caput('xx.VAL',3.0,wait=True)
    """
    start_time = time.time()
    thispv = get_pv(pvname, timeout=timeout, connect=True)
    if thispv.connected:
        timeout -= (time.time() - start_time)
        return thispv.put(value, wait=wait, timeout=timeout)

def caget(pvname, as_string=False, count=None, as_numpy=True,
          use_monitor=False, timeout=5.0):
    """caget(pvname, as_string=False,count=None,as_numpy=True,
             use_monitor=False,timeout=5.0)
    simple get of a pv's value..
       >>> x = caget('xx.VAL')

    to get the character string representation (formatted double,
    enum string, etc):
       >>> x = caget('xx.VAL', as_string=True)

    to get a truncated amount of data from an array, you can specify
    the count with
       >>> x = caget('MyArray.VAL', count=1000)
    """
    start_time = time.time()
    thispv = get_pv(pvname, timeout=timeout, connect=True)
    if thispv.connected:
        if as_string:
            thispv.get_ctrlvars()
        timeout -= (time.time() - start_time)
        val = thispv.get(count=count, timeout=timeout,
                         use_monitor=use_monitor,
                         as_string=as_string,
                         as_numpy=as_numpy)
        poll()
        return val

def cainfo(pvname, print_out=True, timeout=5.0):
    """cainfo(pvname,print_out=True,timeout=5.0)

    return printable information about pv
       >>>cainfo('xx.VAL')

    will return a status report for the pv.

    If print_out=False, the status report will be printed,
    and not returned.
    """
    start_time = time.time()
    thispv = get_pv(pvname, timeout=timeout, connect=True)
    if thispv.connected:
        conn_time = time.time() - start_time
        thispv.get(timeout=timeout-conn_time)
        get_time = time.time() - start_time
        thispv.get_ctrlvars(timeout=timeout-get_time)
        if print_out:
            ca.write(thispv.info)
        else:
            return thispv.info

def camonitor_clear(pvname):
    """clear a monitor on a PV"""
    if pvname in _PVmonitors_:
        _PVmonitors_[pvname].remove_callback(index=-999)
        _PVmonitors_.pop(pvname)

def camonitor(pvname, writer=None, callback=None):
    """ camonitor(pvname, writer=None, callback=None)

    sets a monitor on a PV.
       >>>camonitor('xx.VAL')

    This will write a message with the latest value for that PV each
    time the value changes and when ca.poll() is called.

    To write the result to a file, provide the writer option a write method
    to an open file or some other method that accepts a string.

    To completely control where the output goes, provide a callback method
    and you can do whatever you'd like with them.

    Your callback will be sent keyword arguments for pvname, value, and
    char_value Important: use **kwd!!
    """

    if writer is None:
        writer = ca.write
    if callback is None:
        def callback(pvname=None, value=None, char_value=None, **kwds):
            "generic monitor callback"
            if char_value is None:
                char_value = repr(value)
            writer("%.32s %s %s" % (pvname, pv.fmt_time(), char_value))

    thispv = get_pv(pvname, connect=True)
    if thispv.connected:
        thispv.get()
        thispv.add_callback(callback, index=-999, with_ctrlvars=True)
        _PVmonitors_[pvname] = thispv

def caget_many(pvlist, as_string=False, count=None, as_numpy=True, timeout=5.0):
    """get values for a list of PVs
    This does not maintain PV objects, and works as fast
    as possible to fetch many values.
    """
    chids, out = [], []
    for name in pvlist: chids.append(ca.create_channel(name,
                                                       auto_cb=False,
                                                       connect=False))
    for chid in chids: ca.connect_channel(chid)
    for chid in chids: ca.get(chid, count=count, as_string=as_string, as_numpy=as_numpy, wait=False)
    for chid in chids: out.append(ca.get_complete(chid,
                                                  count=count,
                                                  as_string=as_string,
                                                  as_numpy=as_numpy,
                                                  timeout=timeout))
    return out

def caput_many(pvlist, values, wait=False, connection_timeout=None, put_timeout=60):
    """put values to a list of PVs, as fast as possible
    This does not maintain the PV objects it makes.  If
    wait is 'each', *each* put operation will block until
    it is complete or until the put_timeout duration expires.
    If wait is 'all', this method will block until *all*
    put operations are complete, or until the put_timeout
    duration expires.
    Note that the behavior of 'wait' only applies to the
    put timeout, not the connection timeout.
    Returns a list of integers for each PV, 1 if the put
    was successful, or a negative number if the timeout
    was exceeded.
    """
    if len(pvlist) != len(values):
        raise ValueError("List of PV names must be equal to list of values.")
    out = []
    pvs = [PV(name, auto_monitor=False, connection_timeout=connection_timeout) for name in pvlist]
    conns = [p.connected for p in pvs]
    wait_all = (wait == 'all')
    wait_each = (wait == 'each')
    for p, v in zip(pvs, values):
        out.append(p.put(v, wait=wait_each, timeout=put_timeout, use_complete=wait_all))
    if wait_all:
        start_time = time.time()
        while not all([(p.connected and p.put_complete) for p in pvs]):
            ca.poll()
            elapsed_time = time.time() - start_time
            if elapsed_time > put_timeout:
                break
        return [1 if (p.connected and p.put_complete) else -1 for p in pvs]
    else:
        return [o if o == 1 else -1 for o in out]





"""
String and data utils, where implementation differs between Python 2 & 3
"""
import sys
from copy import deepcopy
import os

PY_MAJOR, PY_MINOR = sys.version_info[:2]

if PY_MAJOR >= 3:
    from . import utils3 as utils_mod
else:
    from . import utils2 as utils_mod

STR2BYTES = utils_mod.STR2BYTES
BYTES2STR = utils_mod.BYTES2STR
NULLCHAR = utils_mod.NULLCHAR
NULLCHAR_2 = utils_mod.NULLCHAR_2
strjoin = utils_mod.strjoin
is_string = utils_mod.is_string
is_string_or_bytes = utils_mod.is_string_or_bytes
ascii_string = utils_mod.ascii_string

memcopy = deepcopy
if PY_MAJOR == 2 and PY_MINOR == 5:
    def memcopy(a):
        return a

def clib_search_path(lib):
    '''Assemble path to c library.

    Parameters
    ----------
    lib : str
        Either 'ca' or 'Com'.

    Returns
    --------
    str : string

    Examples
    --------
    >>> clib_search_path('ca')
    'linux64/libca.so'

    '''

    # determine which libca / libCom dll is appropriate
    try:
        import platform
        nbits = platform.architecture()[0]
        mach = platform.machine()
    except:
        nbits = '32bit'
        mach = 'x86_64'

    nbits = nbits.replace('bit', '')
    if mach.startswith('arm'):
        nbits = 'arm'

    libfmt = 'lib%s.so'
    if os.name == 'nt':
        libsrc = 'win'
        libfmt = '%s.dll'
    elif sys.platform == 'darwin':
        libsrc = 'darwin'
        libfmt = 'lib%s.dylib'
    elif sys.platform.startswith('linux'):
        libsrc = 'linux'

    else:
        return None

    return os.path.join("%s%s" % (libsrc, nbits), libfmt % lib)



#!/usr/bin/env python
# unit-tests for ca interface

import sys
import time
import unittest
import numpy
import threading
import pytest

from contextlib import contextmanager
from epics import PV, get_pv, caput, caget, caget_many, caput_many, ca

import pvnames

def write(msg):
    sys.stdout.write(msg)
    sys.stdout.flush()

CONN_DAT ={}
CHANGE_DAT = {}

def onConnect(pvname=None, conn=None, chid=None,  **kws):
    write('  :Connection status changed:  %s  connected=%s\n' % (pvname, conn))
    global CONN_DAT
    CONN_DAT[pvname] = conn

def onChanges(pvname=None, value=None, **kws):
    write( '/// New Value: %s  value=%s, kw=%s\n' %( pvname, str(value), repr(kws)))
    global CHANGE_DAT
    CHANGE_DAT[pvname] = value

@contextmanager
def no_simulator_updates():
    '''Context manager which pauses and resumes simulator PV updating'''
    try:
        caput(pvnames.pause_pv, 1, wait=True)
        time.sleep(0.05)
        yield
    finally:
        caput(pvnames.pause_pv, 0, wait=True)


class PV_Tests(unittest.TestCase):
    def testA_CreatePV(self):
        write('Simple Test: create pv\n')
        pv = get_pv(pvnames.double_pv)
        self.assertIsNot(pv, None)

    def testA_CreatedWithConn(self):
        write('Simple Test: create pv with conn callback\n')
        pv = get_pv(pvnames.int_pv, connection_callback=onConnect)
        val = pv.get()

        global CONN_DAT
        conn = CONN_DAT.get(pvnames.int_pv, None)
        self.assertEqual(conn, True)

    def test_caget(self):
        write('Simple Test of caget() function\n')
        pvs = (pvnames.double_pv, pvnames.enum_pv, pvnames.str_pv)
        for p in pvs:
            val = caget(p)
            self.assertIsNot(val, None)
        sval = caget(pvnames.str_pv)
        self.assertEqual(sval, 'ao')

    def test_caget_many(self):
        write('Simple Test of caget_many() function\n')
        pvs = [pvnames.double_pv, pvnames.enum_pv, pvnames.str_pv]
        vals = caget_many(pvs)
        self.assertEqual(len(vals), len(pvs))
        self.assertIsInstance(vals[0], float)
        self.assertIsInstance(vals[1], int)
        self.assertIsInstance(vals[2], str)

    def test_caput_many_wait_all(self):
        write('Test of caput_many() function, waiting for all.\n')
        pvs = [pvnames.double_pv, pvnames.enum_pv, 'ceci nest pas une PV']
        #pvs = ["MTEST:Val1", "MTEST:Val2", "MTEST:SlowVal"]
        vals = [0.5, 0, 23]
        t0 = time.time()
        success = caput_many(pvs, vals, wait='all', connection_timeout=0.5, put_timeout=5.0)
        t1 = time.time()
        self.assertEqual(len(success), len(pvs))
        self.assertEqual(success[0], 1)
        self.assertEqual(success[1], 1)
        self.failUnless(success[2] < 0)


    def test_caput_many_wait_each(self):
        write('Simple Test of caput_many() function, waiting for each.\n')
        pvs = [pvnames.double_pv, pvnames.enum_pv, 'ceci nest pas une PV']
        #pvs = ["MTEST:Val1", "MTEST:Val2", "MTEST:SlowVal"]
        vals = [0.5, 0, 23]
        success = caput_many(pvs, vals, wait='each', connection_timeout=0.5, put_timeout=1.0)
        self.assertEqual(len(success), len(pvs))
        self.assertEqual(success[0], 1)
        self.assertEqual(success[1], 1)
        self.failUnless(success[2] < 0)

    def test_caput_many_no_wait(self):
        write('Simple Test of caput_many() function, without waiting.\n')
        pvs = [pvnames.double_pv, pvnames.enum_pv, 'ceci nest pas une PV']
        #pvs = ["MTEST:Val1", "MTEST:Val2", "MTEST:SlowVal"]
        vals = [0.5, 0, 23]
        success = caput_many(pvs, vals, wait=None, connection_timeout=0.5)
        self.assertEqual(len(success), len(pvs))
        #If you don't wait, ca.put returns 1 as long as the PV connects
        #and the put request is valid.
        self.assertEqual(success[0], 1)
        self.assertEqual(success[1], 1)
        self.failUnless(success[2] < 0)

    def test_get1(self):
        write('Simple Test: test value and char_value on an integer\n')
        with no_simulator_updates():
            pv = get_pv(pvnames.int_pv)
            val = pv.get()
            cval = pv.get(as_string=True)

            self.failUnless(int(cval)== val)

    def test_get_with_metadata(self):
        with no_simulator_updates():
            pv = get_pv(pvnames.int_pv, form='native')

            # Request time type
            md = pv.get_with_metadata(use_monitor=False, form='time')
            assert 'timestamp' in md
            assert 'lower_ctrl_limit' not in md

            # Request control type
            md = pv.get_with_metadata(use_monitor=False, form='ctrl')
            assert 'lower_ctrl_limit' in md
            assert 'timestamp' not in md

            # Use monitor: all metadata should come through
            md = pv.get_with_metadata(use_monitor=True)
            assert 'timestamp' in md
            assert 'lower_ctrl_limit' in md

            # Get a namespace
            ns = pv.get_with_metadata(use_monitor=True, as_namespace=True)
            assert hasattr(ns, 'timestamp')
            assert hasattr(ns, 'lower_ctrl_limit')

    def test_get_string_waveform(self):
        write('String Array: \n')
        with no_simulator_updates():
            pv = get_pv(pvnames.string_arr_pv)
            val = pv.get()
            self.failUnless(len(val) > 10)
            self.assertIsInstance(val[0], str)
            self.failUnless(len(val[0]) > 1)
            self.assertIsInstance(val[1], str)
            self.failUnless(len(val[1]) > 1)

    def test_putcomplete(self):
        write('Put with wait and put_complete (using real motor!) \n')
        vals = (1.35, 1.50, 1.44, 1.445, 1.45, 1.453, 1.446, 1.447, 1.450, 1.450, 1.490, 1.5, 1.500)
        p = get_pv(pvnames.motor1)
        # this works with a real motor, fail if it doesn't connect quickly
        if not p.wait_for_connection(timeout=0.2):
            self.skipTest('Unable to connect to real motor record')

        see_complete = []
        for v in vals:
            t0 = time.time()
            p.put(v, use_complete=True)
            count = 0
            for i in range(100000):
                time.sleep(0.001)
                count = count + 1
                if p.put_complete:
                    see_complete.append(True)
                    break
                # print( 'made it to value= %.3f, elapsed time= %.4f sec (count=%i)' % (v, time.time()-t0, count))
        self.failUnless(len(see_complete) > (len(vals) - 5))

    def test_putwait(self):
        write('Put with wait (using real motor!) \n')
        pv = get_pv(pvnames.motor1)
        # this works with a real motor, fail if it doesn't connect quickly
        if not pv.wait_for_connection(timeout=0.2):
            self.skipTest('Unable to connect to real motor record')

        val = pv.get()

        t0 = time.time()
        if val < 5:
            pv.put(val + 1.0, wait=True)
        else:
            pv.put(val - 1.0, wait=True)
        dt = time.time()-t0
        write('    put took %s sec\n' % dt)
        self.failUnless( dt > 0.1)

        # now with a callback!
        global put_callback_called
        put_callback_called = False

        def onPutdone(pvname=None, **kws):
            print( 'put done ', pvname, kws)
            global put_callback_called
            put_callback_called = True
        val = pv.get()
        if  val < 5:
            pv.put(val + 1.0, callback=onPutdone)
        else:
            pv.put(val - 1.0, callback=onPutdone)

        t0 = time.time()
        while time.time()-t0 < dt*1.50:
            time.sleep(0.02)

        write('    put should be done by now?  %s \n' % put_callback_called)
        self.failUnless(put_callback_called)

        # now using pv.put_complete
        val = pv.get()
        if  val < 5:
            pv.put(val + 1.0, use_complete=True)
        else:
            pv.put(val - 1.0, use_complete=True)
        t0 = time.time()
        count = 0
        while time.time()-t0 < dt*1.50:
            if pv.put_complete:
                break
            count = count + 1
            time.sleep(0.02)
        write('    put_complete=%s (should be True), and count=%i (should be>3)\n' %
              (pv.put_complete, count))
        self.failUnless(pv.put_complete)
        self.failUnless(count > 3)

    def test_get_callback(self):
        write("Callback test:  changing PV must be updated\n")
        global NEWVALS
        mypv = get_pv(pvnames.updating_pv1)
        NEWVALS = []
        def onChanges(pvname=None, value=None, char_value=None, **kw):
            write( 'PV %s %s, %s Changed!\n' % (pvname, repr(value), char_value))
            NEWVALS.append( repr(value))

        mypv.add_callback(onChanges)
        write('Added a callback.  Now wait for changes...\n')

        t0 = time.time()
        while time.time() - t0 < 3:
            time.sleep(1.e-4)
        write('   saw %i changes.\n' % len(NEWVALS))
        self.failUnless(len(NEWVALS) > 3)
        mypv.clear_callbacks()

    def test_put_string_waveform(self):
        write('String Array: put\n')
        with no_simulator_updates():
            pv = get_pv(pvnames.string_arr_pv)
            put_value = ['a', 'b', 'c']
            pv.put(put_value, wait=True)
            get_value = pv.get(use_monitor=False, as_numpy=False)
            numpy.testing.assert_array_equal(get_value, put_value)

    def test_put_string_waveform_single_element(self):
        write('String Array: put single element\n')
        with no_simulator_updates():
            pv = get_pv(pvnames.string_arr_pv)
            put_value = ['a']
            pv.put(put_value, wait=True)
            time.sleep(0.05)
            get_value = pv.get(use_monitor=False, as_numpy=False)
            self.failUnless(put_value[0] == get_value)

    def test_put_string_waveform_mixed_types(self):
        write('String Array: put mixed types\n')
        with no_simulator_updates():
            pv = get_pv(pvnames.string_arr_pv)
            put_value = ['a', 2, 'b']
            pv.put(put_value, wait=True)
            time.sleep(0.05)
            get_value = pv.get(use_monitor=False, as_numpy=False)
            numpy.testing.assert_array_equal(get_value, ['a', '2', 'b'])

    def test_put_string_waveform_empty_list(self):
        write('String Array: put empty list\n')
        with no_simulator_updates():
            pv = get_pv(pvnames.string_arr_pv)
            put_value = []
            pv.put(put_value, wait=True)
            time.sleep(0.05)
            get_value = pv.get(use_monitor=False, as_numpy=False)
            self.failUnless('' == ''.join(get_value))

    def test_put_string_waveform_zero_length_strings(self):
        write('String Array: put zero length strings\n')
        with no_simulator_updates():
            pv = get_pv(pvnames.string_arr_pv)
            put_value = ['', '', '']
            pv.put(put_value, wait=True)
            time.sleep(0.05)
            get_value = pv.get(use_monitor=False, as_numpy=False)
            numpy.testing.assert_array_equal(get_value, put_value)

    def test_subarrays(self):
        write("Subarray test:  dynamic length arrays\n")
        driver = get_pv(pvnames.subarr_driver)
        subarr1 = get_pv(pvnames.subarr1)
        subarr1.connect()

        len_full = 64
        len_sub1 = 16
        full_data = numpy.arange(len_full)/1.0

        caput("%s.NELM" % pvnames.subarr1, len_sub1)
        caput("%s.INDX" % pvnames.subarr1, 0)


        driver.put(full_data) ;
        time.sleep(0.1)
        subval = subarr1.get()

        self.assertEqual(len(subval), len_sub1)
        self.failUnless(numpy.all(subval == full_data[:len_sub1]))
        write("Subarray test:  C\n")
        caput("%s.NELM" % pvnames.subarr2, 19)
        caput("%s.INDX" % pvnames.subarr2, 3)

        subarr2 = get_pv(pvnames.subarr2)
        subarr2.get()

        driver.put(full_data) ;   time.sleep(0.1)
        subval = subarr2.get()

        self.assertEqual(len(subval), 19)
        self.failUnless(numpy.all(subval == full_data[3:3+19]))

        caput("%s.NELM" % pvnames.subarr2, 5)
        caput("%s.INDX" % pvnames.subarr2, 13)

        driver.put(full_data) ;   time.sleep(0.1)
        subval = subarr2.get()

        self.assertEqual(len(subval), 5)
        self.failUnless(numpy.all(subval == full_data[13:5+13]))

    def test_subarray_zerolen(self):
        subarr1 = get_pv(pvnames.zero_len_subarr1)
        subarr1.wait_for_connection()

        val = subarr1.get(use_monitor=True, as_numpy=True)
        self.assertIsInstance(val, numpy.ndarray, msg='using monitor')
        self.assertEquals(len(val), 0, msg='using monitor')
        self.assertEquals(val.dtype, numpy.float64, msg='using monitor')

        val = subarr1.get(use_monitor=False, as_numpy=True)
        self.assertIsInstance(val, numpy.ndarray, msg='no monitor')
        self.assertEquals(len(val), 0, msg='no monitor')
        self.assertEquals(val.dtype, numpy.float64, msg='no monitor')


    def test_waveform_get_with_count_arg(self):
        with no_simulator_updates():
            # NOTE: do not use get_pv() here, as `count` is incompatible with
            # the cache
            wf = PV(pvnames.char_arr_pv, count=32)
            val=wf.get()
            self.assertEquals(len(val), 32)

            val=wf.get(count=wf.nelm)
            self.assertEquals(len(val), wf.nelm)


    def test_waveform_callback_with_count_arg(self):
        values = []

        # NOTE: do not use get_pv() here, as `count` is incompatible with
        # the cache
        wf = PV(pvnames.char_arr_pv, count=32)
        def onChanges(pvname=None, value=None, char_value=None, **kw):
            write( 'PV %s %s, %s Changed!\n' % (pvname, repr(value), char_value))
            values.append( value)

        wf.add_callback(onChanges)
        write('Added a callback.  Now wait for changes...\n')

        t0 = time.time()
        while time.time() - t0 < 3:
            time.sleep(1.e-4)
            if len(values)>0:
                break

        self.failUnless(len(values) > 0)
        self.assertEquals(len(values[0]),32)

        wf.clear_callbacks()



    def test_emptyish_char_waveform_no_monitor(self):
        '''a test of a char waveform of length 1 (NORD=1): value "\0"
        without using auto_monitor
        '''
        with no_simulator_updates():
            zerostr = PV(pvnames.char_arr_pv, auto_monitor=False)
            zerostr.wait_for_connection()

            # elem_count = 128, requested count = None, libca returns count = 1
            zerostr.put([0], wait=True)
            self.assertEquals(zerostr.get(as_string=True), '')
            numpy.testing.assert_array_equal(zerostr.get(as_string=False), [0])
            self.assertEquals(zerostr.get(as_string=True, as_numpy=False), '')
            numpy.testing.assert_array_equal(zerostr.get(as_string=False, as_numpy=False), [0])

            # elem_count = 128, requested count = None, libca returns count = 2
            zerostr.put([0, 0], wait=True)
            self.assertEquals(zerostr.get(as_string=True), '')
            numpy.testing.assert_array_equal(zerostr.get(as_string=False), [0, 0])
            self.assertEquals(zerostr.get(as_string=True, as_numpy=False), '')
            numpy.testing.assert_array_equal(zerostr.get(as_string=False, as_numpy=False), [0, 0])
            zerostr.disconnect()

    def test_emptyish_char_waveform_monitor(self):
        '''a test of a char waveform of length 1 (NORD=1): value "\0"
        with using auto_monitor
        '''
        with no_simulator_updates():
            zerostr = PV(pvnames.char_arr_pv, auto_monitor=True)
            zerostr.wait_for_connection()

            zerostr.put([0], wait=True)
            time.sleep(0.2)

            self.assertEquals(zerostr.get(as_string=True), '')
            numpy.testing.assert_array_equal(zerostr.get(as_string=False), [0])
            self.assertEquals(zerostr.get(as_string=True, as_numpy=False), '')
            numpy.testing.assert_array_equal(zerostr.get(as_string=False, as_numpy=False), [0])

            zerostr.put([0, 0], wait=True)
            time.sleep(0.2)

            self.assertEquals(zerostr.get(as_string=True), '')
            numpy.testing.assert_array_equal(zerostr.get(as_string=False), [0, 0])
            self.assertEquals(zerostr.get(as_string=True, as_numpy=False), '')
            numpy.testing.assert_array_equal(zerostr.get(as_string=False, as_numpy=False), [0, 0])
            zerostr.disconnect()

    def testEnumPut(self):
        pv = get_pv(pvnames.enum_pv)
        self.assertIsNot(pv, None)
        pv.put('Stop')
        time.sleep(0.1)
        val = pv.get()
        self.assertEqual(val, 0)


    def test_DoubleVal(self):
        pvn = pvnames.double_pv
        pv = get_pv(pvn)
        pv.get()
        cdict  = pv.get_ctrlvars()
        write( 'Testing CTRL Values for a Double (%s)\n'   % (pvn))
        self.failUnless('severity' in cdict)
        self.failUnless(len(pv.host) > 1)
        self.assertEqual(pv.count,1)
        self.assertEqual(pv.precision, pvnames.double_pv_prec)
        units= ca.BYTES2STR(pv.units)
        self.assertEqual(units, pvnames.double_pv_units)
        self.failUnless(pv.access.startswith('read'))


    def test_type_converions_2(self):
        write("CA type conversions arrays\n")
        pvlist = (pvnames.char_arr_pv,
                  pvnames.long_arr_pv,
                  pvnames.double_arr_pv)
        with no_simulator_updates():
            chids = []
            for name in pvlist:
                chid = ca.create_channel(name)
                ca.connect_channel(chid)
                chids.append((chid, name))
                ca.poll(evt=0.025, iot=5.0)
            ca.poll(evt=0.05, iot=10.0)

            values = {}
            for chid, name in chids:
                values[name] = ca.get(chid)
            for promotion in ('ctrl', 'time'):
                for chid, pvname in chids:
                    write('=== %s  chid=%s as %s\n' % (ca.name(chid),
                                                       repr(chid), promotion))
                    time.sleep(0.01)
                    if promotion == 'ctrl':
                        ntype = ca.promote_type(chid, use_ctrl=True)
                    else:
                        ntype = ca.promote_type(chid, use_time=True)

                    val  = ca.get(chid, ftype=ntype)
                    cval = ca.get(chid, as_string=True)
                    for a, b in zip(val, values[pvname]):
                        self.assertEqual(a, b)

    def test_waveform_get_1elem(self):
        pv = get_pv(pvnames.double_arr_pv)
        val = pv.get(count=1, use_monitor=False)
        self.failUnless(isinstance(val, numpy.ndarray))
        self.failUnless(len(val), 1)

    def test_subarray_1elem(self):
        with no_simulator_updates():
            # pv = get_pv(pvnames.zero_len_subarr1)
            pv = get_pv(pvnames.double_arr_pv)
            pv.wait_for_connection()

            val = pv.get(count=1, use_monitor=False)
            print('val is', val, type(val))
            self.assertIsInstance(val, numpy.ndarray)
            self.assertEqual(len(val), 1)

            val = pv.get(count=1, as_numpy=False, use_monitor=False)
            print('val is', val, type(val))
            self.assertIsInstance(val, list)
            self.assertEqual(len(val), 1)


@pytest.mark.parametrize('num_threads', [1, 10, 200])
@pytest.mark.parametrize('thread_class', [ca.CAThread, threading.Thread])
def test_multithreaded_get(num_threads, thread_class):
    def thread(thread_idx):
        result[thread_idx] = (pv.get(),
                              pv.get_with_metadata(form='ctrl')['value'],
                              pv.get_with_metadata(form='time')['value'],
                              )

    result = {}
    ca.use_initial_context()
    pv = get_pv(pvnames.double_pv)

    threads = [thread_class(target=thread, args=(i, ))
               for i in range(num_threads)]

    with no_simulator_updates():
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

    assert len(result) == num_threads
    print(result)
    values = set(result.values())
    assert len(values) == 1

    value, = values
    assert value is not None


@pytest.mark.parametrize('num_threads', [1, 10, 100])
def test_multithreaded_put_complete(num_threads):
    def callback(pvname, data):
        result.append(data)

    def thread(thread_idx):
        pv.put(thread_idx, callback=callback,
               callback_data=dict(data=thread_idx),
               wait=True)
        time.sleep(0.1)

    result = []
    ca.use_initial_context()
    pv = get_pv(pvnames.double_pv)

    threads = [ca.CAThread(target=thread, args=(i, ))
               for i in range(num_threads)]

    with no_simulator_updates():
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

    assert len(result) == num_threads
    print(result)
    assert set(result) == set(range(num_threads))


def test_force_connect():
    pv = get_pv(pvnames.double_arrays[0], auto_monitor=True)

    print("Connecting")
    assert pv.wait_for_connection(5.0)

    print("SUM", pv.get().sum())

    time.sleep(3)

    print("Disconnecting")
    pv.disconnect()
    print("Reconnecting")

    pv.force_connect()
    assert pv.wait_for_connection(5.0)

    called = {'called': False}

    def callback(value=None, **kwargs):
        called['called'] = True
        print("update", value.sum())

    pv.add_callback(callback)

    time.sleep(1)
    assert pv.get() is not None
    assert called['called']


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase( PV_Tests)
    unittest.TextTestRunner(verbosity=1).run(suite)


#     chid = ca.create_channel(pvnames.int_pv,
#                              callback=onConnect)
#
#     time.sleep(0.1)



from __future__ import print_function
import sys
from contextlib import contextmanager
import epics
import multiprocessing as mp
import pytest

import pvnames
PVS = [pvnames.double_pv, pvnames.double_pv2]


@contextmanager
def pool_ctx():
    pool = epics.CAPool()
    yield pool
    pool.close()
    pool.join()


@pytest.mark.skipif(sys.version_info >= (3, 0),
                    reason='CAPool not functioning in Python 3')
def test_caget():
    with pool_ctx() as pool:
        print('Using caget() in subprocess pools:')
        print('\tpool.process =', pool.Process)
        values = pool.map(epics.caget, PVS)

        for pv, value in zip(PVS, values):
            print('\t%s = %s' % (pv, value))


def _manager_test_fcn(pv_dict, pv):
    pv_dict[pv] = epics.caget(pv)


@pytest.mark.skipif(sys.version_info >= (3, 0),
                    reason='CAPool not functioning in Python 3')
def test_manager():
    '''
    Fill up a shared dictionary using a manager
    '''
    with pool_ctx() as pool:
        print('Multiprocessing Manager test:')

        manager = mp.Manager()
        pv_dict = manager.dict()

        results = [pool.apply_async(_manager_test_fcn, [pv_dict, pv])
                   for pv in PVS]

    print('\tResulting pv dictionary: %s' % pv_dict)



from ctypes import cdll
from epics.ca import find_libca
dllname = find_libca()

print(dllname)

libca = cdll.LoadLibrary(dllname)
print("{:20s} {:s} {:s}".format('function name', 'in libca' , 'in None'))

for fname in ('ca_context_create', 'ca_pend_io',
                  'ca_create_channel'):

    print("{:20s}  {:s}     {:s}".format(fname,
                                         repr(hasattr(libca, fname)),
                                         repr(hasattr(cdll.LoadLibrary(None), fname))))

#
# False
# >>> libca.ca_context_create
# <_FuncPtr object at 0x109536c00>
# >>> hasattr(cdll.LoadLibrary(None), 'sqrt')
# True
# >>> libc = cdll.LoadLibrary('libc')
# Traceback (most recent call last):
#   File "<stdin>", line 1, in <module>
#   File "/Users/Newville/anaconda3/lib/python3.7/ctypes/__init__.py", line 434, in LoadLibrary
#     return self._dlltype(name)
#   File "/Users/Newville/anaconda3/lib/python3.7/ctypes/__init__.py", line 356, in __init__
#     self._handle = _dlopen(self._name, mode)
# OSError: dlopen(libc, 6): image not found
# >>> libc = cdll.LoadLibrary('libc.dylib')
# >>> hasattr(cdll.LoadLibrary(None), 'sqrt')
# True
# >>> libca.ca_context_create
# <_FuncPtr object at 0x109536c00>
# >>> hasattr(cdll.LoadLibrary(None), 'ca_context_create')
# False
# >>> hasattr(cdll.LoadLibrary(None), 'ca_pend_io')
# False
# >>> hasattr(cdll.LoadLibrary(None), 'ca_pend_event')
# False
# >>> libca.ca_pend_event
# <_FuncPtr object at 0x109536e58>
# >>>



#!/usr/bin/env python
# from distutils.core import setup
from setuptools import setup

import os
import sys
import epics
import shutil

import versioneer

long_desc = '''Python Interface to the Epics Channel Access protocol
of the Epics control system.   PyEpics provides 3 layers of access to
Channel Access (CA):

1. a light wrapping of the CA C library calls, using ctypes. This
   provides a procedural CA library in which the user is expected
   to manage Channel IDs. It is mostly provided as a foundation
   upon which higher-level access is built.
2. PV() (Process Variable) objects, which represent the basic object
   in CA, allowing one to keep a persistent connection to a remote
   Process Variable.
3. A simple set of functions caget(), caput() and so on to mimic
   the CA command-line tools and give the simplest access to CA.

In addition, the library includes convenience classes to define
Devices -- collections of PVs that might represent an Epics Record
or physical device (say, a camera, amplifier, or power supply), and
to help write GUIs for CA.
'''

#
no_libca="""*******************************************************
*** WARNING - WARNING - WARNING - WARNING - WARNING ***

       Could not find CA dynamic library!

A dynamic library (libca.so or libca.dylib) for EPICS CA
must be found in order for EPICS calls to work properly.

Please read the INSTALL inststructions, and fix this
problem before tyring to use the epics package.
*******************************************************
"""

nolibca = os.environ.get('NOLIBCA', None)
if nolibca is None:
    pkg_data = {'epics.clibs': ['darwin64/*', 'linux64/*', 'linux32/*',
                                'linuxarm/*', 'win32/*', 'win64/*']}
else:
    pkg_data = dict()

PY_MAJOR, PY_MINOR = sys.version_info[:2]
if PY_MAJOR == 2 and PY_MINOR < 6:
    shutil.copy(pjoin('epics', 'utils3.py'),
                pjoin('epics', 'utils3_save_py.txt'))
    shutil.copy(pjoin('epics', 'utils2.py'),
                pjoin('epics', 'utils3.py'))

setup(name = 'pyepics',
      version = versioneer.get_version(),
      cmdclass = versioneer.get_cmdclass(),
      author       = 'Matthew Newville',
      author_email = 'newville@cars.uchicago.edu',
      url          = 'http://pyepics.github.io/pyepics/',
      download_url = 'http://pyepics.github.io/pyepics/',
      license      = 'Epics Open License',
      description  = "Epics Channel Access for Python",
      long_description = long_desc,
      platforms    = ['Windows', 'Linux', 'Mac OS X'],
      classifiers  = ['Intended Audience :: Science/Research',
                      'Operating System :: OS Independent',
                      'Programming Language :: Python',
                      'Topic :: Scientific/Engineering'],
      packages = ['epics','epics.wx','epics.devices', 'epics.compat',
                  'epics.autosave', 'epics.clibs'],
      package_data = pkg_data,
     )

try:
    libca = epics.ca.find_libca()
except:
    sys.stdout.write("%s" % no_libca)



import time
import epics
import pvnames
pvname = pvnames.double_pv

mypv = epics.PV(pvname)


def wait(timeout=6):
    t0 = time.time()
    while time.time() - t0 < timeout:
        time.sleep(1.e-3)
        epics.poll()


print( 'Created PV = ', mypv)
def CB1(pvname=None, value=None, char_value=None, **kw):
    print( 'CB1 PV Changed! ', pvname, value, char_value )


def CB2(pvname=None, value=None, char_value=None, **kw):
    print( 'CB2 ! ', pvname, value, char_value)

mypv.add_callback(CB1)

print( 'Added a callback.  Now wait for changes')

print( 'ready')

wait(10)
print( 'now, add another: ')
mypv.add_callback(CB2)

wait(10)






#!/usr/bin/env python
# test expression parsing

import os
import sys
import time
import unittest
import numpy
import ctypes
from contextlib import contextmanager
from epics import ca, dbr, caput

import pvnames

def _ca_connect(chid,timeout=5.0):
    n  = 0
    t0 = time.time()
    conn = 2==ca.state(chid)
    while (not conn) and (time.time()-t0 < timeout):
        ca.poll(1.e-6,1.e-4)
        conn = 2==ca.state(chid)
        n += 1
    return conn, time.time()-t0, n

def write(msg):
    sys.stdout.write('%s\n'% msg)
    sys.stdout.flush()

CONN_DAT ={}
CHANGE_DAT = {}

def onConnect(pvname=None, conn=None, chid=None,  **kws):
    write('  /// Connection status changed:  %s  %s' % (pvname, repr(kws)))
    global CONN_DAT
    CONN_DAT[pvname] = conn

def onChanges(pvname=None, value=None, **kws):
    write( '/// New Value: %s  value=%s, kw=%s' %( pvname, str(value), repr(kws)))
    global CHANGE_DAT
    CHANGE_DAT[pvname] = value

@contextmanager
def no_simulator_updates():
    '''Context manager which pauses and resumes simulator PV updating'''
    try:
        caput(pvnames.pause_pv, 1)
        yield
    finally:
        caput(pvnames.pause_pv, 0)

class CA_BasicTests(unittest.TestCase):

    def setUp(self):
        write('Starting %s...' % self.id().split(".")[-1])
    def tearDown(self):
        write('Completed %s...\n' % self.id().split(".")[-1])

    def testA_CreateChid(self):
        write('Simple Test: create chid')
        chid = ca.create_channel(pvnames.double_pv)
        self.assertIsNot(chid, None)

    def testA_GetNonExistentPV(self):
        write('Simple Test: get on a non-existent PV')
        chid = ca.create_channel('Definitely-Not-A-Real-PV')
        val = ca.get(chid)
        self.assertEqual(val, None)

    def testA_CreateChid_CheckTypeCount(self):
        write('Simple Test: create chid, check count, type, host, and access')
        chid = ca.create_channel(pvnames.double_pv)
        ret = ca.connect_channel(chid)
        ca.pend_event(1.e-3)

        ftype  = ca.field_type(chid)
        count  = ca.element_count(chid)
        host    = ca.host_name(chid)
        rwacc = ca.access(chid)

        self.assertIsNot(chid, None)
        self.assertIsNot(host, None)
        self.assertEqual(count, 1)
        self.assertEqual(ftype, 6)
        self.assertEqual(rwacc,'read/write')


    def testA_CreateChidWithConn(self):
        write('Simple Test: create chid with conn callback')
        chid = ca.create_channel(pvnames.int_pv,
                                 callback=onConnect)
        val = ca.get(chid)

        global CONN_DAT
        conn = CONN_DAT.get(pvnames.int_pv, None)
        self.assertEqual(conn, True)

    def test_dbrName(self):
        write( 'DBR Type Check')
        self.assertEqual(dbr.Name(dbr.STRING), 'STRING')
        self.assertEqual(dbr.Name(dbr.FLOAT),  'FLOAT')
        self.assertEqual(dbr.Name(dbr.ENUM), 'ENUM')
        self.assertEqual(dbr.Name(dbr.CTRL_CHAR), 'CTRL_CHAR')
        self.assertEqual(dbr.Name(dbr.TIME_DOUBLE), 'TIME_DOUBLE')
        self.assertEqual(dbr.Name(dbr.TIME_LONG), 'TIME_LONG')

        self.assertEqual(dbr.Name('STRING', reverse=True), dbr.STRING)
        self.assertEqual(dbr.Name('DOUBLE', reverse=True), dbr.DOUBLE)
        self.assertEqual(dbr.Name('CTRL_ENUM', reverse=True), dbr.CTRL_ENUM)
        self.assertEqual(dbr.Name('TIME_LONG', reverse=True), dbr.TIME_LONG)

    def test_Connect1(self):
        chid = ca.create_channel(pvnames.double_pv)
        conn,dt,n = _ca_connect(chid, timeout=2)
        write( 'CA Connection Test1: connect to existing PV')
        write( ' connected in %.4f sec' % (dt))
        self.assertEqual(conn,True)

    def test_Connected(self):
        pvn = pvnames.double_pv
        chid = ca.create_channel(pvn,connect=True)
        isconn = ca.isConnected(chid)
        write( 'CA test Connected (%s) = %s' % (pvn,isconn))
        self.assertEqual(isconn,True)
        state= ca.state(chid)
        self.assertEqual(state,ca.dbr.CS_CONN)
        acc = ca.access(chid)
        self.assertEqual(acc,'read/write')


    def test_DoubleVal(self):
        pvn = pvnames.double_pv
        chid = ca.create_channel(pvn,connect=True)
        cdict  = ca.get_ctrlvars(chid)
        write( 'CA testing CTRL Values for a Double (%s)'   % (pvn))
        self.failUnless('units' in cdict)
        self.failUnless('precision' in cdict)
        self.failUnless('severity' in cdict)

        hostname = ca.host_name(chid)
        self.failUnless(len(hostname) > 1)

        count = ca.element_count(chid)
        self.assertEqual(count,1)

        ftype= ca.field_type(chid)
        self.assertEqual(ftype,ca.dbr.DOUBLE)

        prec = ca.get_precision(chid)
        self.assertEqual(prec, pvnames.double_pv_prec)

        units= ca.BYTES2STR(ca.get_ctrlvars(chid)['units'])
        self.assertEqual(units, pvnames.double_pv_units)

        rwacc= ca.access(chid)
        self.failUnless(rwacc.startswith('read'))


    def test_UnConnected(self):
        write( 'CA Connection Test1: connect to non-existing PV (2sec timeout)')
        chid = ca.create_channel('impossible_pvname_certain_to_fail')
        conn,dt,n = _ca_connect(chid, timeout=2)
        self.assertEqual(conn,False)


    def test_putwait(self):
        'test put with wait'
        pvn = pvnames.non_updating_pv
        chid = ca.create_channel(pvn, connect=True)
        o  = ca.put(chid, -1, wait=True)
        val = ca.get(chid)
        self.assertEqual(val, -1)
        o  = ca.put(chid, 2, wait=True)
        val = ca.get(chid)
        self.assertEqual(val, 2)

    def test_promote_type(self):
        pvn = pvnames.double_pv
        chid = ca.create_channel(pvn,connect=True)
        write( 'CA promote type (%s)' % (pvn))
        f_t  = ca.promote_type(chid,use_time=True)
        f_c  = ca.promote_type(chid,use_ctrl=True)
        self.assertEqual(f_t, ca.dbr.TIME_DOUBLE)
        self.assertEqual(f_c, ca.dbr.CTRL_DOUBLE)

    def test_ProcPut(self):
        pvn  = pvnames.enum_pv
        chid = ca.create_channel(pvn, connect=True)
        write( 'CA test put to PROC Field (%s)' % (pvn))
        for input in (1, '1', 2, '2', 0, '0', 50, 1):
            ret = None
            try:
                ret = ca.put(chid, 1)
            except:
                pass
            self.assertIsNot(ret, None)

    def test_subscription_double(self):
        pvn = pvnames.updating_pv1
        chid = ca.create_channel(pvn,connect=True)
        cb, uarg, eventID = ca.create_subscription(chid, callback=onChanges)

        start_time = time.time()
        global CHANGE_DAT
        while time.time()-start_time < 5.0:
            time.sleep(0.01)
            if CHANGE_DAT.get(pvn, None) is not None:
                break
        val = CHANGE_DAT.get(pvn, None)
        ca.clear_subscription(eventID)
        self.assertIsNot(val, None)

    def test_subscription_custom(self):
        pvn = pvnames.updating_pv1
        chid = ca.create_channel(pvn, connect=True)

        global change_count
        change_count = 0

        def my_callback(pvname=None, value=None, **kws):
            write( ' Custom Callback  %s  value=%s' %(pvname, str(value)))
            global change_count
            change_count = change_count + 1

        cb, uarg, eventID = ca.create_subscription(chid, callback=my_callback)

        start_time = time.time()
        while time.time()-start_time < 2.0:
            time.sleep(0.01)

        ca.clear_subscription(eventID)
        time.sleep(0.2)
        self.assertTrue(change_count > 2)

    def test_subscription_str(self):

        pvn = pvnames.updating_str1
        write(" Subscription on string: %s " % pvn)
        chid = ca.create_channel(pvn,connect=True)
        cb, uarg, eventID = ca.create_subscription(chid, callback=onChanges)

        start_time = time.time()
        global CHANGE_DAT
        while time.time()-start_time < 3.0:
            time.sleep(0.01)
            ca.put(chid, "%.1f" % (time.time()-start_time) )
            if CHANGE_DAT.get(pvn, None) is not None:
                break
        val = CHANGE_DAT.get(pvn, None)
        # ca.clear_subscription(eventID)
        self.assertIsNot(val, None)
        time.sleep(0.2)


    def _test_array_callback(self, arrayname, array_type, length, element_type):
        """ Helper function to subscribe to a PV array and check it
        receives at least one subscription callback w/ specified type,
        length & uniform element type. Checks separately for normal,
        TIME & CTRL subscription variants. Returns the array or fails
        an assertion."""
        results = {}
        for form in [ 'normal', 'time', 'ctrl' ]:
            chid = ca.create_channel(arrayname,connect=True)
            cb, uarg, eventID = ca.create_subscription(chid, use_time=form=='time', use_ctrl=form=='ctrl', callback=onChanges)

            CHANGE_DAT.pop(arrayname, None)
            timeout=0
            # wait up to 6 seconds, if no callback probably due to simulator.py
            # not running...
            while timeout<120 and not arrayname in CHANGE_DAT:
                time.sleep(0.05)
                timeout = timeout+1
            val = CHANGE_DAT.get(arrayname, None)
            ca.clear_subscription(eventID)
            self.assertIsNot(val, None)
            self.assertEqual(type(val), array_type)
            self.assertEqual(len(val), length)
            self.assertEqual(type(val[0]), element_type)
            self.assertTrue(all( type(e)==element_type for e in val))
            results[form] = val
        return results

    def test_subscription_long_array(self):
        """ Check that numeric arrays callbacks successfully send correct data """
        self._test_array_callback(pvnames.long_arr_pv, numpy.ndarray, 2048, numpy.int32)

    def test_subscription_double_array(self):
        """ Check that double arrays callbacks successfully send correct data """
        self._test_array_callback(pvnames.double_arr_pv, numpy.ndarray, 2048, numpy.float64)

    def test_subscription_string_array(self):
        """ Check that string array callbacks successfully send correct data """
        results = self._test_array_callback(pvnames.string_arr_pv, list, 128, str)
        self.assertTrue(len(results["normal"][0]) > 0)
        self.assertTrue(len(results["time"][0]) > 0)
        self.assertTrue(len(results["ctrl"][0]) > 0)

    def test_subscription_char_array(self):
        """ Check that uchar array callbacks successfully send correct data as arrays """
        self._test_array_callback(pvnames.char_arr_pv, numpy.ndarray, 128, numpy.uint8)



    def test_Values(self):
        write( 'CA test Values (compare 5 values with caget)')
        os.system('rm ./caget.tst')
        vals = {}
        with no_simulator_updates():
            for pvn in (pvnames.str_pv,  pvnames.int_pv,
                        pvnames.float_pv, pvnames.enum_pv,
                        pvnames.long_pv):
                os.system('caget  -n -f5 %s >> ./caget.tst' % pvn)
                chid = ca.create_channel(pvn)
                ca.connect_channel(chid)
                vals[pvn] = ca.get(chid)
            rlines = open('./caget.tst', 'r').readlines()
            for line in rlines:
                pvn, sval = [i.strip() for i in line[:-1].split(' ', 1)]
                tval = str(vals[pvn])
                if pvn in (pvnames.float_pv,pvnames.double_pv):
                    # use float precision!
                    tval = "%.5f" % vals[pvn]
                self.assertEqual(tval, sval)

    def test_type_converions_1(self):
        write("CA type conversions scalars")
        pvlist = (pvnames.str_pv, pvnames.int_pv, pvnames.float_pv,
                  pvnames.enum_pv,  pvnames.long_pv,  pvnames.double_pv2)
        chids = []
        with no_simulator_updates():
            for name in pvlist:
                chid = ca.create_channel(name)
                ca.connect_channel(chid)
                chids.append((chid, name))
                ca.poll(evt=0.025, iot=5.0)
            ca.poll(evt=0.05, iot=10.0)

            values = {}
            for chid, name in chids:
                values[name] = ca.get(chid, as_string=True)

            for promotion in ('ctrl', 'time'):
                for chid, pvname in chids:
                    write('=== %s  chid=%s as %s' % (ca.name(chid), repr(chid),
                                                     promotion))
                    time.sleep(0.01)
                    if promotion == 'ctrl':
                        ntype = ca.promote_type(chid, use_ctrl=True)
                    else:
                        ntype = ca.promote_type(chid, use_time=True)

                    val  = ca.get(chid, ftype=ntype)
                    cval = ca.get(chid, as_string=True)
                    if ca.element_count(chid) > 1:
                        val = val[:12]
                    self.assertEqual(cval, values[pvname])

    def test_type_converions_2(self):
        write("CA type conversions arrays")
        pvlist = (pvnames.char_arr_pv,
                  pvnames.long_arr_pv,
                  pvnames.double_arr_pv)
        with no_simulator_updates():
            chids = []
            for name in pvlist:
                chid = ca.create_channel(name)
                ca.connect_channel(chid)
                chids.append((chid, name))
                ca.poll(evt=0.025, iot=5.0)
            ca.poll(evt=0.05, iot=10.0)

            values = {}
            for chid, name in chids:
                values[name] = ca.get(chid)
            for promotion in ('ctrl', 'time'):
                for chid, pvname in chids:
                    write('=== %s  chid=%s as %s' % (ca.name(chid), repr(chid),
                                                     promotion))
                    time.sleep(0.01)
                    if promotion == 'ctrl':
                        ntype = ca.promote_type(chid, use_ctrl=True)
                    else:
                        ntype = ca.promote_type(chid, use_time=True)

                    val  = ca.get(chid, ftype=ntype)
                    cval = ca.get(chid, as_string=True)
                    for a, b in zip(val, values[pvname]):
                        self.assertEqual(a, b)


    def test_Array0(self):
        write('Array Test: get double array as numpy array, ctypes Array, and list')
        chid = ca.create_channel(pvnames.double_arrays[0])
        aval = ca.get(chid)
        cval = ca.get(chid, as_numpy=False)

        self.assertTrue(isinstance(aval, numpy.ndarray))
        self.assertTrue(len(aval) > 2)

        self.assertTrue(isinstance(cval, ctypes.Array))
        self.assertTrue(len(cval) > 2)
        lval = list(cval)
        self.assertTrue(isinstance(lval, list))
        self.assertTrue(len(lval) > 2)
        self.assertTrue(lval == list(aval))

    def test_xArray1(self):
        write('Array Test: get(wait=False) / get_complete()')
        chid = ca.create_channel(pvnames.double_arrays[0])
        val0 = ca.get(chid)
        aval = ca.get(chid, wait=False)
        self.assertTrue(aval is  None)
        val1 = ca.get_complete(chid)
        self.assertTrue(all(val0 == val1))

    def test_xArray2(self):
        write('Array Test: get fewer than max vals using ca.get(count=0)')
        chid = ca.create_channel(pvnames.double_arrays[0])
        maxpts = ca.element_count(chid)
        npts = int(max(2, maxpts/2.3 - 1))
        write('max points is %s' % (maxpts, ))
        dat = numpy.random.normal(size=npts)
        write('setting array to a length of npts=%s' % (npts, ))
        ca.put(chid, dat)
        out1 = ca.get(chid)
        self.assertTrue(isinstance(out1, numpy.ndarray))
        self.assertEqual(len(out1), npts)
        out2 = ca.get(chid, count=0)
        self.assertTrue(isinstance(out2, numpy.ndarray))
        self.assertEqual(len(out2), npts)

    def test_xArray3(self):
        write('Array Test: get char array as string')
        chid = ca.create_channel(pvnames.char_arrays[0])
        val = ca.get(chid)
        self.assertTrue(isinstance(val, numpy.ndarray))
        char_val = ca.get(chid, as_string=True)
        self.assertTrue(isinstance(char_val, str))
        conv = ''.join([chr(i) for i in val])
        self.assertEqual(conv, char_val)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase( CA_BasicTests)
    unittest.TextTestRunner(verbosity=1).run(suite)


#     chid = ca.create_channel(pvnames.int_pv,
#                              callback=onConnect)
#
#     time.sleep(0.1)
# ;



import time
import epics
import sys
import pvnames

pvname = pvnames.updating_pv1 # motor1

mypv = epics.PV(pvname)

write = sys.stdout.write

mypv.get_ctrlvars()

write('Created PV = %s\n' % mypv)
def onChanges(pvname=None, value=None, char_value=None, **kw):
    write( 'PV %s %s, %s Changed!\n' % (pvname, repr(value), char_value))

mypv.add_callback(onChanges)

write('Added a callback.  Now wait for changes...\n')

def wait(timeout=10):
    t0 = time.time()
    while time.time() - t0 < timeout: time.sleep(1.e-4)

wait(10)
